@@ -1,15 +1,21 @@
 from __future__ import unicode_literals
 
-import hmac
 import binascii
 import struct
-from hashlib import sha256
+import json
+
 from base64 import standard_b64encode, urlsafe_b64decode, urlsafe_b64encode
 
 from libnacl.secret import SecretBox
 
 from macaroons.caveat import Caveat
-from macaroons.utils import truncate_or_pad, convert_to_bytes
+from macaroons.utils import (truncate_or_pad,
+                             convert_to_bytes,
+                             convert_to_string,
+                             generate_derived_key,
+                             hmac_digest,
+                             sign_first_party_caveat,
+                             sign_third_party_caveat)
 
 
 class RawMacaroon(object):
@@ -28,8 +34,9 @@ class RawMacaroon(object):
                  identifier=None,
                  key=None,
                  serialized=None,
-                 signature=None):
-        self._serialized = serialized
+                 signature=None,
+                 json=None):
+        self._serialized = None
         self._caveats = []
         if location and identifier and key:
             self._location = location
@@ -47,6 +54,8 @@ class RawMacaroon(object):
         elif serialized:
             self._serialized = serialized
             self._deserialize(serialized)
+        elif json:
+            self._deserialize_json(json)
         else:
             pass
 
@@ -93,7 +102,13 @@ class RawMacaroon(object):
         return urlsafe_b64encode(combined)
 
     def serialize_json(self):
-        pass
+        serialized = {
+            'location': convert_to_string(self.location),
+            'identifier': convert_to_string(self.identifier),
+            'caveats': [caveat.to_dict() for caveat in self.caveats],
+            'signature': convert_to_string(self.signature)
+        }
+        return json.dumps(serialized)
 
     def _deserialize(self, data):
         PACKET_PREFIX_LENGTH = 4
@@ -146,6 +161,20 @@ class RawMacaroon(object):
 
         return self
 
+    def _deserialize_json(self, data):
+        deserialized = json.loads(convert_to_string(data))
+        self._location = deserialized['location']
+        self._identifier = deserialized['identifier']
+        for c in deserialized['caveats']:
+            caveat = Caveat(
+                caveatId=c['cid'],
+                verificationKeyId=c['vid'],
+                location=c['cl']
+            )
+            self._caveats.append(caveat)
+        self._signature = convert_to_bytes(deserialized['signature'])
+        return self
+
     def inspect(self):
         combined = 'location' + ' ' + self.location + '\n'
         combined += 'identifier' + ' ' + self.identifier + '\n'
@@ -155,10 +184,10 @@ class RawMacaroon(object):
 
             if caveat.verificationKeyId and caveat.location:
                 combined += 'vid' + ' ' + \
-                    caveat.verificationKeyId.decode('ascii') + '\n'
+                    caveat._verificationKeyId.decode('ascii') + '\n'
                 combined += 'cl' + ' ' + caveat.location + '\n'
 
-        combined += 'signature' + ' ' + self.signature.decode('ascii')
+        combined += 'signature' + ' ' + self._signature.decode('ascii')
         return combined
 
     # TODO
@@ -175,11 +204,7 @@ class RawMacaroon(object):
     # the wrong location by binding to the root macaroon
     def prepare_for_request(self, macaroon):
         protected = macaroon.copy()
-        protected._signature = self._macaroon_hmac_concat(
-            b'\0',
-            self.signature,
-            macaroon.signature
-        )
+        protected._signature = self._bind_signature(macaroon.signature)
         return protected
 
     # The existing macaroon signature is the key for hashing the
@@ -189,7 +214,7 @@ class RawMacaroon(object):
         caveat = Caveat(caveatId=predicate)
         self._caveats.append(caveat)
         encode_key = binascii.unhexlify(self.signature)
-        self._signature = self._macaroon_hmac(encode_key, predicate)
+        self._signature = sign_first_party_caveat(encode_key, predicate)
         return self
 
     # The third party caveat key is encrypted useing the current signature, and
@@ -197,7 +222,7 @@ class RawMacaroon(object):
     # is the key for hashing the string (verificationId + caveatId).
     # This new hash becomes the signature of the macaroon with caveat added.
     def add_third_party_caveat(self, location, key, key_id):
-        derived_key = truncate_or_pad(self._generate_derived_key(key))
+        derived_key = truncate_or_pad(generate_derived_key(key))
         self._add_third_party_caveat_direct(
             location,
             derived_key,
@@ -211,7 +236,7 @@ class RawMacaroon(object):
                                        key,
                                        key_id,
                                        nonce=None):
-        old_key = truncate_or_pad(self.signature)
+        old_key = truncate_or_pad(binascii.unhexlify(self.signature))
         box = SecretBox(key=old_key)
         encrypted = box.encrypt(key, nonce=nonce)
         verificationKeyId = standard_b64encode(encrypted)
@@ -222,55 +247,41 @@ class RawMacaroon(object):
         )
         self._caveats.append(caveat)
         encode_key = binascii.unhexlify(self.signature)
-        self._signature = self._macaroon_hmac_concat(
+        self._signature = sign_third_party_caveat(
             encode_key,
             caveat._verificationKeyId,
             caveat._caveatId
         )
         return self
 
+    def _bind_signature(self, signature):
+        bound = self._macaroon_hmac_concat(
+            truncate_or_pad(b'0'),
+            self.signature,
+            signature
+        )
+        return bound
+
     # Hashes two strings, then concatenates them and hashes the combined
     # string
     def _macaroon_hmac_concat(self, key, data1, data2):
-        hash1 = hmac.new(
-            key,
-            msg=data1,
-            digestmod=sha256
-        ).digest()
-        hash2 = hmac.new(
-            key,
-            msg=data2,
-            digestmod=sha256
-        ).digest()
+        hash1 = hmac_digest(key, binascii.unhexlify(data1))
+        hash2 = hmac_digest(key, binascii.unhexlify(data2))
         combined = hash1 + hash2
-        return hmac.new(
-            key,
-            msg=combined,
-            digestmod=sha256
-        ).hexdigest().encode('ascii')
+        return self._macaroon_hmac(
+            key, combined
+        )
 
     # Given a high-entropy root key _key and an identifier id, this returns
     # a valid signature sig = MAC(k, id).
     def _create_initial_macaroon_signature(self, key):
-        derived_key = self._generate_derived_key(key)
+        derived_key = generate_derived_key(key)
         return self._macaroon_hmac(derived_key, self._identifier)
 
-    def _generate_derived_key(self, key):
-        generator_key = b'macaroons-key-generator'
-        derived_key = self._hmac(generator_key, key)
-        return derived_key
-
     def _macaroon_hmac(self, key, data):
-        dig = self._hmac(key, data)
+        dig = hmac_digest(key, data)
         return binascii.hexlify(dig)
 
-    def _hmac(self, key, data):
-        return hmac.new(
-            key,
-            msg=data,
-            digestmod=sha256
-        ).digest()
-
     def _packetize(self, key, data):
         PACKET_PREFIX_LENGTH = 4
         # The 2 covers the space and the newline
